"""git constants"""

from typing import Dict

GitConfig = Dict[str, str]

HEAD: str = 'HEAD'
ORIGIN: str = 'origin'
FETCH_URL: str = 'fetch_url'
PUSH_URL: str = 'push_url'
GITMODULES: str = '.gitmodules'
